// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "dark_gen.h"

#include "linalg.h"

#include <ranges>

namespace spex {

// Generate offset correponding to short exposure times
static auto genOffsetShort(const SPEXSettings& settings,
                           const std::vector<L1A>& l1a_products,
                           CKD& ckd) -> void
{
    ckd.dark.offset_short.resize(ckd.npix, 0.0);
    int n_l1a_short {};
    for (const auto& l1a :
         l1a_products
           | std::views::filter(
             [short_exposure_time =
                settings.dark.short_exposure_time](const auto& l1a) {
                 return l1a.exposure_time < short_exposure_time;
             })) {
        spdlog::info(
          "Generating short exposure offset from exposure time {:0.5f} ms",
          1e3 * l1a.exposure_time);
        for (int i {}; i < ckd.npix; ++i) {
            ckd.dark.offset_short[i] += l1a.image[i];
        }
        ++n_l1a_short;
    }
    if (n_l1a_short == 0) {
        throw std::runtime_error { "make sure that at least one measurement "
                                   "has exposure time shorter than "
                                   + std::to_string(
                                     settings.dark.short_exposure_time)
                                   + " s [dark][short_exposure_time]" };
    }
    for (int i {}; i < ckd.npix; ++i) {
        ckd.dark.offset_short[i] /= n_l1a_short;
    }
}

// Generate offset correponding to long exposure times and dark current
static auto genOffsetLongAndCurrent(
  const SPEXSettings& settings,
  const std::vector<L1A>::const_iterator l1a_products_beg,
  const std::vector<L1A>::const_iterator l1a_products_end,
  CKD& ckd) -> void
{
    ckd.dark.offset_long.resize(ckd.npix);
    ckd.dark.current.resize(ckd.npix);
    ckd.dark.chi2.resize(ckd.npix);
    ckd.dark.sigma_beta.resize(ckd.npix);
    ckd.dark.sigma_alpha.resize(ckd.npix);
    ckd.dark.last_exposure_times.resize(ckd.npix);
    // Only use measurements longer than the long exposure time and
    // shorter than the maximum allowed exposure time beyond which the
    // signal becomes saturated.
    const std::vector<std::reference_wrapper<const L1A>> l1a_refs {
        std::find_if(
          l1a_products_beg,
          l1a_products_end,
          [exposure_time = settings.dark.long_exposure_time](const auto& l1a) {
              return l1a.exposure_time > exposure_time;
          }),
        l1a_products_end
    };
    spdlog::info(
      "Generating long exposure offsets and dark current from exposure times:");
    for (const auto& l1a : l1a_refs) {
        spdlog::info("  {:2.5f} s", l1a.get().exposure_time);
    }
    // Default is to include all long exposure times in the set
    int n_l1a { static_cast<int>(l1a_refs.size()) };
    // Optionally reduce the number of files by an exposure time cutoff
    if (settings.dark.last_exposure_time) {
        n_l1a = static_cast<int>(std::ranges::distance(
          l1a_refs.begin(),
          std::ranges::find_if(
            l1a_refs,
            [exposure_time =
               settings.dark.last_exposure_time](const auto& l1a) {
                return l1a.get().exposure_time > exposure_time;
            })));
    }
    // At least this many exposure times are always includes in the linear fit
    const int n_l1a_min { static_cast<int>(settings.dark.data_start_fraction
                                           * n_l1a) };
    std::vector<double> exposure_times(n_l1a);
#pragma omp parallel for
    for (int i_pix = 0; i_pix < ckd.npix; ++i_pix) {
        // Fit for S = offset_long + tau * dark_current where tau is
        // the exposure time using least squares percentage
        // regression.
        double Q_00 {};
        double Q_01 {};
        double Q_02 {};
        double Q_11 {};
        double Q_12 {};
        double Q_22 {};
        for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
            exposure_times[i_l1a] = l1a_refs[i_l1a].get().exposure_time;
        }
        // Starting uncertainty of the dark current
        ckd.dark.sigma_alpha[i_pix] = 1e100;
        for (int i_l1a {}; i_l1a < n_l1a; ++i_l1a) {
            // Use weights to effectively spread the exposure times
            // evenly over the whole range, making the result less
            // sensitive to the actual spacing.
            double w;
            if (i_l1a == 0) {
                w = 2 * (exposure_times[1] - exposure_times.front());
            } else if (i_l1a == n_l1a - 1) {
                w = 2 * (exposure_times.back() - exposure_times[n_l1a - 2]);
            } else {
                w = exposure_times[i_l1a + 1] - exposure_times[i_l1a - 1];
            }
            const double w_over_exp { w / exposure_times[i_l1a] };
            const double w_over_exp2 { w_over_exp / exposure_times[i_l1a] };
            Q_00 += w;
            Q_01 += w_over_exp;
            Q_02 += w_over_exp2;
            const double& counts { l1a_refs[i_l1a].get().image[i_pix] };
            Q_11 += w_over_exp * counts;
            Q_12 += w_over_exp2 * counts;
            Q_22 += w_over_exp2 * counts * counts;
            const double Z_1 { Q_00 * Q_02 - Q_01 * Q_01 };
            const double Z_2 { Q_00 * Q_12 - Q_01 * Q_11 };
            const double Z_3 { Q_02 * Q_11 - Q_01 * Q_12 };
            const double offset_long { Z_2 / Z_1 };
            const double current { Z_3 / Z_1 };
            const double chi2 { Q_22 - Q_12 * offset_long - Q_11 * current };
            const double sigma_beta { std::sqrt(chi2 / (n_l1a - 1) * Q_00
                                                / Z_1) };
            const double sigma_alpha { std::sqrt(chi2 / (n_l1a - 1) * Q_02
                                                 / Z_1) };
            // The final values of the dark offset, current, and
            // related parameters are determined by either the last
            // exposure time specified by the user or the set of
            // exposure times yielding the smaller dark current
            // uncertainty. In the latter case, if the current
            // uncertainty is smaller than the smallest one yet
            // encountered for this pixel, stored in
            // ckd.dark.sigma_alpha, then update that and other CKD
            // variables.
            if (i_l1a > n_l1a_min
                && (settings.dark.last_exposure_time
                    || sigma_alpha < ckd.dark.sigma_alpha[i_pix])) {
                ckd.dark.offset_long[i_pix] = offset_long;
                ckd.dark.current[i_pix] = current;
                ckd.dark.chi2[i_pix] = chi2;
                ckd.dark.sigma_beta[i_pix] = sigma_beta;
                ckd.dark.sigma_alpha[i_pix] = sigma_alpha;
                ckd.dark.last_exposure_times[i_pix] = exposure_times[i_l1a];
            }
        }
    }
}

// Generate temperature dependent dark current
static auto genTempDepCurrent(const ProcLevel proc_level,
                              SPEXSettings settings,
                              CKD& ckd) -> void
{
    if (settings.dark.l1a_files_T.empty()) {
        return;
    }
    // L1A files for the temperature dependence are handled by an L1A
    // manager. First copy the file list to settings.l1a_files because
    // that's where the L1A manager will look for.
    settings.l1a_files[static_cast<int>(ProcLevel::dark)] =
      static_cast<std::vector<std::string>>(settings.dark.l1a_files_T);
    L1ADiagnostic l1a_manager { false };
    l1a_manager.init(settings, proc_level, ckd);
    // Only the trivial binning table can be an option here
    std::shared_ptr<BinningTable> binning_table {
        std::make_shared<BinningTable>(
          settings.main.detector_spat / settings.main.bin[0],
          settings.main.detector_spec / settings.main.bin[1],
          "",
          0,
          false)
    };
    spdlog::info("Reading L1A files for temperature dependent dark current");
    for (L1A& l1a : l1a_manager.l1a_products) {
        l1a.binning_table = binning_table;
        l1a.readImage(ckd);
    }
    // Temperature threshold for sorting
    constexpr double temp_thres { 1.0 };
    std::ranges::sort(
      l1a_manager.l1a_products, [](const auto& l1a_1, const auto& l1a_2) {
          if (std::abs(l1a_1.detector_temperature - l1a_2.detector_temperature)
              > temp_thres) {
              return l1a_1.detector_temperature < l1a_2.detector_temperature;
          }
          return l1a_1.exposure_time < l1a_2.exposure_time;
      });
    // Detect all temperatures
    std::vector<double> temperatures {
        l1a_manager.l1a_products.front().detector_temperature
    };
    for (const auto& l1a : l1a_manager.l1a_products) {
        if (std::abs(l1a.detector_temperature - temperatures.back())
            > temp_thres) {
            temperatures.push_back(l1a.detector_temperature);
        }
    }
    // Dark currents corresponding to different temperatures
    std::vector<std::vector<double>> currents {};
    for (const double& temperature : temperatures) {
        spdlog::info("Using measurements at temperature {:8.4f} K",
                     temperature);
        // Find a sublist that corresponds to the given temperature
        auto l1a_first { std::ranges::find_if(
          l1a_manager.l1a_products, [temperature](const auto& l1a) {
              return std::abs(l1a.detector_temperature - temperature)
                     < temp_thres;
          }) };
        auto l1a_last { std::ranges::find_if(
          l1a_first,
          l1a_manager.l1a_products.end(),
          [temperature](const auto& l1a) {
              return std::abs(l1a.detector_temperature - temperature)
                     > temp_thres;
          }) };
        genOffsetLongAndCurrent(settings, l1a_first, l1a_last, ckd);
        // We save dark current but otherwise everything in ckd.dark
        // will be overwritten by repeated calls to
        // genOffsetLongAndCurrent.
        currents.push_back(ckd.dark.current);
    }
    // Dark currents at different temperatures for a given pixel
    std::vector<double> current(temperatures.size());
    double intersect {}; // dummy variable for the linear regression
    for (int i_pix {}; i_pix < ckd.npix; ++i_pix) {
        if (ckd.pixel_mask[i_pix]) {
            continue;
        }
        for (int i {}; i < static_cast<int>(temperatures.size()); ++i) {
            current[i] = currents[i][i_pix];
        }
        simpleLinearRegression(
          temperatures, current, ckd.dark.current_slope[i_pix], intersect);
    }
}

static auto updateMask(const SPEXSettings& settings, CKD& ckd) -> void
{
    if (settings.dark.mask_offset_min) {
        updateMaskLt(
          ckd.dark.offset_short,
          settings.dark.mask_offset_min.value(),
          "pixels masked bad because their offset (short) is less than",
          ckd.pixel_mask);
        updateMaskLt(
          ckd.dark.offset_long,
          settings.dark.mask_offset_min.value(),
          "pixels masked bad because their offset (long) is less than",
          ckd.pixel_mask);
    }
    if (settings.dark.mask_offset_max) {
        updateMaskGt(
          ckd.dark.offset_short,
          settings.dark.mask_offset_max.value(),
          "pixels masked bad because their offset (short) is greater than",
          ckd.pixel_mask);
        updateMaskGt(
          ckd.dark.offset_long,
          settings.dark.mask_offset_max.value(),
          "pixels masked bad because their offset (long) is greater than",
          ckd.pixel_mask);
    }
    if (settings.dark.mask_current_min) {
        updateMaskLt(
          ckd.dark.current,
          settings.dark.mask_current_min.value(),
          "pixels masked bad because their dark current is less than",
          ckd.pixel_mask);
    }
    if (settings.dark.mask_current_max) {
        updateMaskGt(
          ckd.dark.current,
          settings.dark.mask_current_max.value(),
          "pixels masked bad because their dark current is greater than",
          ckd.pixel_mask);
    }
}

auto DarkGen::processBatch(const int /*i_batch*/) -> void
{
    // Sort all measurements by exposure time because that is the main
    // criterion for which ones to select
    std::ranges::sort(batches.front().l1a_products,
                      [](const auto& l1a_1, const auto& l1a_2) {
                          return l1a_1.exposure_time < l1a_2.exposure_time;
                      });
    genOffsetShort(settings, batches.front().l1a_products, ckd);
    ckd.dark.reference_temperature = settings.dark.reference_temperature;
    ckd.dark.current_slope.resize(ckd.npix, 0.0);
    genTempDepCurrent(proc_level, settings, ckd);
    genOffsetLongAndCurrent(settings,
                            batches.front().l1a_products.cbegin(),
                            batches.front().l1a_products.cend(),
                            ckd);

    // Update pixel mask
    updateMask(settings, ckd);
}

} // namespace spex
