// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "isrf.h"

#include "arb_special_functions.h"
#include "ckd.h"
#include "cubic_spline.h"
#include "linear_spline.h"
#include "spectra.h"

#include <algorithm>
#include <array>
#include <hdf5.h>

namespace spex {

struct parms_type
{
    std::uint16_t mps {};
    std::uint16_t row {};
    std::uint16_t col_mn {};
    std::uint16_t col_mx {};
    float gse_wavelength {};
    float gse_linewidth {};
    float gse_radiance {};
    float rms {};
    double loc {};
    double intgr {};
    double scale_a {};
    double shape_m {};
    double skew {};
    double w_b {};
};

ISRF::ISRF(const std::string& filename, const CKD& ckd)
{
    // Number of rows in a full-frame detector image
    constexpr int n_rows_full { 2048 };
    scale_a.resize(n_rows_full);
    shape_m.resize(n_rows_full);
    skew.resize(n_rows_full);
    radius_r.resize(n_rows_full);
    std::vector<std::vector<double>> col_indices(n_rows_full);

    // Wavelength to column spline in the middle of swaths
    std::vector<std::unique_ptr<CubicSpline>> wave_to_col(ckd.n_vp);
    for (int i_vp {}; i_vp < ckd.n_vp; ++i_vp) {
        const int i_fov { ckd.fov.start_vp[i_vp] + ckd.fov.n_act_vp[i_vp] / 2 };
        std::vector<double> col_indices(ckd.n_detector_cols);
        std::iota(col_indices.begin(), col_indices.end(), 0.0);
        wave_to_col[i_vp] =
          std::make_unique<CubicSpline>(ckd.wave.common[i_fov], col_indices);
    }

    // Read ISRF parameters per viewport from an HDF5 file
    const hid_t file { H5Fopen(filename.c_str(), H5F_ACC_RDONLY, H5P_DEFAULT) };
    const std::vector<std::string> group_names {
        "alt-50.000", "alt-20.000", "alt+00.000", "alt+20.000", "alt+50.000"
    };
    for (int i_vp {}; i_vp < static_cast<int>(group_names.size()); ++i_vp) {
        const std::string dset_name { "/" + group_names[i_vp] + "/issf_parms" };
        const hid_t dset { H5Dopen(file, dset_name.c_str(), H5P_DEFAULT) };
        std::array<hsize_t, 2> dims {};
        H5Sget_simple_extent_dims(H5Dget_space(dset), dims.data(), NULL);
        std::vector<parms_type> parms(dims[0] * dims[1]);
        H5Dread(
          dset, H5Dget_type(dset), H5S_ALL, H5S_ALL, H5P_DEFAULT, parms.data());
        H5Dclose(dset);
        // We don't know in advance how many data points there are per
        // detector row. Resize rows to add more data as it's coming
        // in.
        for (const auto& par_set : parms) {
            const double col { wave_to_col[i_vp]->eval(
              par_set.gse_wavelength) };
            // Skip if this row/column combinations has been detected
            // before (repeaated measurement).
            if (std::ranges::find_if(
                  col_indices[par_set.row],
                  [col](const double x) { return std::abs(col - x) < 1e-10; })
                == col_indices[par_set.row].cend()) {
                col_indices[par_set.row].push_back(col);
                // Flip definition of the scaling parameter
                scale_a[par_set.row].push_back(-par_set.scale_a);
                shape_m[par_set.row].push_back(par_set.shape_m);
                skew[par_set.row].push_back(par_set.skew);
                // We use radius and not the full width
                radius_r[par_set.row].push_back(par_set.w_b / 2);
            }
        }
    }
    H5Fclose(file);

    // The ISRF parameters have been read for each detector row where
    // each row contains 0 or more columns. Next interpolate data
    // across columns such that the each ISRF parameter has the shape
    // of the detector. Note: the number of rows still corresponds to
    // a full-frame image and might have to be binned later.
    // Pivoting and buffer arrays for sorting data by column index
    std::vector<int> piv {};
    std::vector<double> buf {};
    for (int i_row {}; i_row < n_rows_full; ++i_row) {
        if (col_indices[i_row].size() <= 1) {
            scale_a[i_row].assign(ckd.n_detector_cols, 0.0);
            shape_m[i_row].assign(ckd.n_detector_cols, 0.0);
            skew[i_row].assign(ckd.n_detector_cols, 0.0);
            radius_r[i_row].assign(ckd.n_detector_cols, 0.0);
            continue;
        }
        piv.resize(col_indices[i_row].size());
        std::iota(piv.begin(), piv.end(), 0);
        std::ranges::sort(piv, [&col_indices, i_row](const int i, const int j) {
            return col_indices[i_row][i] < col_indices[i_row][j];
        });
        buf.resize(col_indices[i_row].size());
        const auto sort { [&buf, &piv](std::vector<double>& data) {
            for (int i {}; i < static_cast<int>(buf.size()); ++i) {
                buf[i] = data[piv[i]];
            }
            std::swap(buf, data);
        } };
        sort(col_indices[i_row]);
        sort(scale_a[i_row]);
        sort(shape_m[i_row]);
        sort(skew[i_row]);
        sort(radius_r[i_row]);
        const auto interpolate { [](const CKD& ckd,
                                    const std::vector<double>& col_indices,
                                    std::vector<double>& data) {
            const LinearSpline spline { col_indices, data };
            data.resize(ckd.n_detector_cols);
            for (int i {}; i < ckd.n_detector_cols; ++i) {
                data[i] = spline.eval(static_cast<double>(i));
            }
        } };
        interpolate(ckd, col_indices[i_row], scale_a[i_row]);
        interpolate(ckd, col_indices[i_row], shape_m[i_row]);
        interpolate(ckd, col_indices[i_row], skew[i_row]);
        interpolate(ckd, col_indices[i_row], radius_r[i_row]);
    }

    // Bin from full-frame to the CKD dimension across rows
    std::vector<double> scale_a_binned(ckd.npix, 0.0);
    std::vector<double> shape_m_binned(ckd.npix, 0.0);
    std::vector<double> skew_binned(ckd.npix, 0.0);
    std::vector<double> radius_r_binned(ckd.npix, 0.0);
    const auto bin { [](const CKD& ckd,
                        const std::vector<std::vector<double>>& data,
                        std::vector<double>& data_binned) {
        const int bin { n_rows_full / ckd.n_detector_rows };
        for (int i {}; i < ckd.n_detector_rows; ++i) {
            for (int j {}; j < ckd.n_detector_rows; ++j) {
                for (int k {}; k < bin; ++k) {
                    data_binned[i * ckd.n_detector_cols + j] +=
                      data[bin * i + k][j];
                }
            }
        }
        const double norm { 1.0 / bin };
        for (int i {}; i < ckd.npix; ++i) {
            data_binned[i] *= norm;
        }
    } };
    bin(ckd, scale_a, scale_a_binned);
    bin(ckd, shape_m, shape_m_binned);
    bin(ckd, skew, skew_binned);
    bin(ckd, radius_r, radius_r_binned);

    const auto extractSpectra { [](const CKD& ckd,
                                   const std::vector<double>& data_binned,
                                   std::vector<std::vector<double>>& data) {
        data.resize(ckd.fov.n_act, std::vector<double>(ckd.n_detector_cols));
        // Standard deviations and pixel mask are dummy variables here
        std::vector<double> stdev(ckd.npix, 1.0);
        std::vector<bool> pixel_mask(ckd.npix, false);
        for (int i_fov {}; i_fov < ckd.fov.n_act; ++i_fov) {
            Spectra spectra {};
            spectra.extractFOV(ckd, data_binned, stdev, pixel_mask, i_fov);
            for (int i {}; i < ckd.n_detector_cols; ++i) {
                // Average over polarization states
                data[i_fov][i] = 0.5
                                 * (spectra.signal[pol_minus][i]
                                    + spectra.signal[pol_plus][i]);
            }
        }
    } };
    extractSpectra(ckd, scale_a_binned, scale_a);
    extractSpectra(ckd, shape_m_binned, shape_m);
    extractSpectra(ckd, skew_binned, skew);
    extractSpectra(ckd, radius_r_binned, radius_r);
}

[[nodiscard]] auto ISRF::eval(const int i_fov,
                              const int i_spec,
                              const double arg) const -> double
{
    return pearsonIV(arg,
                     scale_a[i_fov][i_spec],
                     shape_m[i_fov][i_spec],
                     skew[i_fov][i_spec],
                     radius_r[i_fov][i_spec]);
}

} // namespace spex
