// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

#include "linear_spline.h"

#include "algorithm.h"
#include "constants.h"

#include <cmath>

namespace spex {

// Threshold for determining if the knots are equally spaced
constexpr double equal_spacing_tol { 1e1
                                     * std::numeric_limits<double>::epsilon() };

auto LinearSpline::setSpacing() -> void
{
    double step0 { knots[1] - knots.front() };
    for (int i { 2 }; i < static_cast<int>(knots.size()); ++i) {
        const double step { knots[i] - knots[i - 1] };
        if (std::abs(step - step0) > equal_spacing_tol) {
            equal_spacing = false;
            break;
        }
    }
    if (equal_spacing) {
        range = knots.back() - knots.front();
    }
}

LinearSpline::LinearSpline(const std::vector<double>& x_values,
                           const std::vector<double>& y_values)
  : knots { x_values }, values { y_values }
{
    setSpacing();
}

auto LinearSpline::lookupIdx(const double x) const -> int
{
    const int idx { static_cast<int>((x - knots.front()) / range
                                     * static_cast<double>(knots.size() - 1)) };
    if (idx < 0 || idx >= static_cast<int>(knots.size())) {
        return def::idx;
    }
    return idx;
}

auto LinearSpline::eval(const double x, const bool is_noise) const -> double
{
    const int idx { equal_spacing ? lookupIdx(x) : binaryFindIdx(knots, x) };
    int i1; // Beginning of subinterval
    int i2; // End of subinterval
    // If x is outside the knot range, extrapolate using the two knots
    // close to either end of the knot sequence.
    if (x <= knots.front()) {
        i1 = 0;
        i2 = 1;
    } else if (x >= knots.back()) {
        i1 = static_cast<int>(knots.size() - 2);
        i2 = static_cast<int>(knots.size() - 1);
    } else {
        i1 = idx;
        i2 = idx + 1;
    }
    const double t { (x - knots[i1]) / (knots[i2] - knots[i1]) };
    if (is_noise) {
        const double a { (knots[i2] - x) / (knots[i2] - knots[i1]) };
        const double b { (x - knots[i1]) / (knots[i2] - knots[i1]) };
        return std::sqrt(a * a * values[i1] * values[i1]
                         + b * b * values[i2] * values[i2]);
    } else {
        return std::lerp(values[i1], values[i2], t);
    }
}

} // namespace spex
