// This source code is licensed under the 3-clause BSD license found
// in the LICENSE file in the root directory of this project.

// Provides a wrapper of netCDF::NcFile that includes some convenience
// member functions and functionality relevant to SPEXone.

#pragma once

#include <netcdf>
#include <string>

namespace spex {

class SPEXSettings;

class NcObject : public netCDF::NcFile
{
private:
    // To keep things simple, the file mode can be set only once
    const netCDF::NcFile::FileMode fmode {};
    std::string filename {};
    bool is_open { false };
    // Use static variables to store some global attributes that are
    // set early on and should not change.
    // The command to invoke this run
    static std::string command_line;
    // L0-L2 processing toolchain version
    static std::string processing_version;

public:
    NcObject(const netCDF::NcFile::FileMode fmode,
             const std::string& filename = {});
    // Call netCDF::NcFile::open using file mode that was set in the
    // constructor. If writing or replacing a file then first always
    // write the global NetCDF attributes of the SPEXone mission.
    auto open(const std::string& filename) -> void;
    auto close() -> void;
    // Whether the NetCDF object is open
    [[nodiscard]] auto isOpen() const -> bool { return is_open; }
    // Re-open with filename defined by an earlier open command. In
    // order to open a different call open with a different name.
    auto reOpen() -> void;
    // Write the contents of the input YAML file into the variable
    // processor_configuration
    auto writeReducedConfig(const SPEXSettings& settings) -> void;
    // Like NcFile::getGroup except can be called with a list of
    // groups names, e.g. getGroup("group1/group2"). There should be
    // no leading slash.
    auto getGroup(const std::string& name_list) const -> netCDF::NcGroup;
    // Copy all attributes from one NetCDF variable to another
    static auto copyAtts(const netCDF::NcVar& in, netCDF::NcVar& out) -> void;
    // Create a copy of a NetCDF variable "var" in group "group". If
    // the dimensions are not given, copy the dimensions and the
    // values from "var". Otherwise, define the new variable with the
    // dimensions "dims" and do not copy the values.
    static auto copyVar(const netCDF::NcVar& var,
                        netCDF::NcGroup& group,
                        const std::vector<netCDF::NcDim>& dims = {})
      -> netCDF::NcVar;
    // Set the global attributes. These are used for any NetCDF
    // product that is produced by this run.
    static auto setGlobalAttributes(const std::string& command_line,
                                    const std::string& processing_version)
      -> void;
    ~NcObject();
};

} // namespace spex
